#!/usr/bin/python -u
#
# Copyright (c) 2008 Citrix Systems, Inc. All use and distribution of this
# copyrighted material is governed by and subject to terms and conditions
# as licensed by Citrix Systems, Inc. All other rights reserved.
# Xen, XenSource and XenEnterprise are either registered trademarks or
# trademarks of Citrix Systems, Inc. in the United States and/or other 
# countries.
#

import xml.dom.minidom
import subprocess, sys, os, glob
import util

HBACLI = "/opt/Citrix/StorageLink/bin/csl_hbascan"
DEVCLI = "/opt/Citrix/StorageLink/bin/csl_devscan"
XML_FORMAT   = "/opt/Citrix/StorageLink/bin/scsi_device_id_config.xml"

def _trace(string):
    """ trace """
    util.SMlog(string)

def is_string(value):
    return isinstance(value, basestring)

def get_single_entry(path):
    f = open(path, 'r')
    line = f.readline()
    f.close()
    return line.rstrip()

def listdir(path):
    cmd = ["ls", path, "-1", "--color=never"]
    try:
        text = pread(cmd)[:-1].split('\n')
    except CommandException, inst:
        if inst.code == errno.ENOENT:
            raise CommandException(errno.EIO, inst.cmd, inst.reason)
        else:
            raise CommandException(inst.code, inst.cmd, inst.reason)
    return text

def doexec(args, inputtext=None):
    """Execute a subprocess, then return its return code, stdout and stderr"""
    proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, 
                                  stderr=subprocess.PIPE, close_fds=True)
    (stdout, stderr) = proc.communicate(inputtext)
    retcode = proc.returncode
    return (retcode, stdout, stderr)

def pread(cmdlist):
    cmdlist_for_exec = []
    for item in cmdlist:
        if is_string(item):
            cmdlist_for_exec.append(item)
        else:
            cmdlist_for_exec.append(item[0])

    (retcode, stdout, stderr) = doexec(cmdlist_for_exec)
    if retcode:
        _trace("zlutil: command FAILED: %s (errno %d) stdout: '%s', stderr: '%s'" % \
                (cmdlist_for_exec, retcode, stdout, stderr))
        raise
    #_trace("command SUCCESS: %s" % cmdlist_for_exec)
    return stdout

def _get_hba_adapters():
    try:
        cmd = [HBACLI, "xml"]
        fc_xml = pread(cmd)
        #strip junk until xml
        xmlstring = fc_xml[fc_xml.index("<HBAInfoList>"):len(fc_xml)]
        xmldoc = xml.dom.minidom.parseString(xmlstring)
        devicedoc = xmldoc.getElementsByTagName("deviceName")
        devices = []
        for dev in devicedoc:
            devices.append(str(dev.firstChild.nodeValue))
        return devices
    except:
        _trace("zlutil: No hba devices detected")
        return []

def _genHostList(procname):
    # loop through and check all adapters
    ids = []
    try:
        for dir in listdir('/sys/class/scsi_host'):
            filename = os.path.join('/sys/class/scsi_host',dir,'proc_name')
            if os.path.exists(filename):
                f = open(filename, 'r')
                if f.readline().find(procname) != -1:
                    ids.append(dir.replace("host",""))
                f.close()
    except:
        pass
    return ids

def _get_iscsi_adapters():
    ids = _genHostList("iscsi_tcp")

    devices = []
    for host in ids:
        try:
            targetIQN = get_single_entry(glob.glob(\
                '/sys/class/iscsi_host/host%s/device/session*/iscsi_session*/targetname' % host)[0])
            addr = get_single_entry(glob.glob(\
                '/sys/class/iscsi_host/host%s/device/session*/connection*/iscsi_connection*/persistent_address' % host)[0])
            port = get_single_entry(glob.glob(\
                '/sys/class/iscsi_host/host%s/device/session*/connection*/iscsi_connection*/persistent_port' % host)[0])
            devices.append("/sys/class/scsi_host/host%s" %host)
        except:
            pass

    return devices

def _rescan(paths, scanstring='- - -'):
    for path in paths:
        _trace("zlutil: RESCANNING %s with %s" % (path, scanstring))
        if os.path.exists(path):
            try:
                f=open(path, 'w')
                f.write('%s\n' % scanstring)
                f.close()
                time.sleep(2)
            except:
                pass

def _zombie_scrub(hba_list, iscsi_list, remove_devices):
    """ check for zombie scsi devices that need to be removed """

    # open and parse scsi proc data
    sfile = open("/proc/scsi/scsi", "r")
    screen = sfile.read()
    sfile.close()

    _trace("zlutil: hba   device list is %s" % hba_list)
    _trace("zlutil: iscsi device list is %s" % iscsi_list)
    #screen scrape! (yucko mooseface)
    # to get scsi devices as understood by the host
    scsi_devices = []
    try:
        while screen:
            # find scsi id
            scsi_idx = screen.index("scsi")
            channel_idx = screen.index(" Channel")
            scsi = screen[scsi_idx:channel_idx]
            scsi_id = scsi.split("scsi")[1].strip()
            # if in list, check for disk vs. controller
            scsi_name = "/sys/class/scsi_host/host%s" % scsi_id
            if (scsi_name in hba_list) or (scsi_name in iscsi_list):
                ltype_idx = screen.index("Type:")
                ltype = screen[ltype_idx+8:ltype_idx+8+13]
                if ltype == "Direct-Access":
                    # extract channel, id, lun
                    elements = screen[channel_idx:channel_idx+27].split(" ")
                    scsi_dev = "%d:%d:%d:%d" % \
                               (int(scsi_id), int(elements[2]), \
                                int(elements[4]), int(elements[6]))
                    scsi_devices.append(scsi_dev)

            # update screen
            screen = screen[scsi_idx:len(screen)]
            screen = screen[screen.index("Host"):len(screen)]
    except:
        pass

    _trace("zlutil: SCSI devices from /proc/scsi/scsi = %s" % scsi_devices)

    # now get devices we can actually touch on the host
    cmd = [DEVCLI, 'native', 'f=%s' % XML_FORMAT, 'xml']
    dev_xml = pread(cmd)

    scsi_luns = []
    try:
        dev_info = xml.dom.minidom.parseString(dev_xml)
        for element in dev_info.firstChild.childNodes:
            if element.nodeType == 1:
                scsi_id = "%s:%s:%s:%s" % \
                    (str(element.getAttribute('adapter')),
                     str(element.getAttribute('bus')),
                     str(element.getAttribute('target')),
                     str(element.getAttribute('lun')))
                scsi_luns.append(scsi_id)
    except:
        raise 

    _trace("zlutil: SCSI luns found by devscan = %s" % scsi_luns)

    # now that we have a list of scsi devices and a list of 
    # scsi luns, run through the list of devices, and if it is 
    # not present in the scsi devices, we will remove it

    found = False
    for dev in scsi_devices:
        if not dev in scsi_luns:
            path = "/sys/class/scsi_device/%s/device/delete" % dev
            found = True
            if (remove_devices == True):
                _trace("zlutil: REMOVING = %s" % path)
                dfile = open(path, 'w')
                dfile.write('1\n')
                dfile.close()
            else:
                _trace("zlutil: FOUND = %s" % path)

    if found == False:
        _trace("zlutil: NO ZOMBIES FOUND")

    if (remove_devices == True):
        _rescan(hba_list)
        _rescan(iscsi_list)

def check_for_zombie_luns(remove_devices = False):
    """ entry for zombie lun check """

    hba_adapters = _get_hba_adapters()
    iscsi_adapters = _get_iscsi_adapters()
    _zombie_scrub(hba_adapters, iscsi_adapters, remove_devices)

